function build_mdl(ascii_intr_file)
%BUILD_MDL builds a CheckMate model from intermediate data file.
%  
%  build_mdl uses the setup and dynamics files already created.
%  build_mdl will create the mdl file for simulink and stateflow 
%  The setup file for creating linearcon object and dynamics file for SCSBlock
%  have already been created. 
%  The model that can read from ascii does not contain events or hiearchical fsm
%  or SCSB. 
%  Note: This function does not handle parameter constraint yet.
%  Note(2): This function can handle only one self loop with reset now.
%  Note(3): This is being upgraded to handle event synchronization between machines.
%  Example
%      build_mdl('ascii_intr_file') (without the .intr extension)
%
%  See also the java program to generate the setup and dynamics file (new_Ascii.java)

input_file_id=fopen([ascii_intr_file '.intr']);
read_a_line=fgetl(input_file_id);

while(read_a_line(1)=='%') read_a_line=fgetl(input_file_id);
end

sys_name=read_a_line;
%disp(['sys_name =' sys_name])

fid=fopen([sys_name '.mdl']);
if fid >= 3
    fclose(fid);  %if system exists then delete it first.
    delete([sys_name '.mdl']);
end

global GLOBAL_APPROX_PARAM;
GLOBAL_APPROX_PARAM=parameters(1);

cmnew=new_system(sys_name);  % create the new system
open_system('template');

read_a_line = fgetl(input_file_id);
if(~strcmp(read_a_line,'NUM_MACHINES'))
   error('Incorrect intermediate file: NUM_MACHINES expected')
end

read_a_line=fgetl(input_file_id);

%1-create the SCSB 
add_block('template/Switched Continuous System', [sys_name '/SCSB']);
m=read_a_line;
set_param([sys_name '/SCSB'], 'nu',m);
%[nx,dummy]=find_node(parse_tree.value, 'variables');
read_a_line = fgetl(input_file_id);
if(~strcmp(read_a_line,'VARIABLES'))
   error('Incorrect intermediate file: VARIABLES expected')
end
read_a_line = fgetl(input_file_id);
set_param([sys_name '/SCSB'], 'swfunc', [sys_name '_dyn']);
%get one initial point 
initial_point=get_initial_point(sys_name);
set_param([sys_name '/SCSB'], 'x0', initial_point, 'nx',read_a_line);
set_param([sys_name '/SCSB'], 'ICS', 'initial_set');
set_param([sys_name '/SCSB'], 'AR', 'analysis_region');

read_a_line = fgetl(input_file_id);
if(~strcmp(read_a_line,'HALFSPACES'))
   error('Incorrect intermediate file: HALFSPACES expected')
end
read_a_line = fgetl(input_file_id);

nhspace = str2num(read_a_line);
%2- create the Polyhedral Thresholds
j=1;
for i=1:nhspace
   %create halfspaces linearcon
   name=fgetl(input_file_id);
   if isempty(strfind(name,'AR')) & isempty(strfind(name,'ICS'))  %don't create pthbs for ICS and AR.
      add_block('template/Polyhedral Threshold', [sys_name '/' name]);
      set_param( [sys_name '/' name], 'polyhedron', name);
      set_param( [sys_name '/' name], 'Position', [ 300, j*75+50, 350,j*75+100]); 
      j=j+1;
      add_line(sys_name, 'SCSB/1', [name '/1'], 'autorouting', 'on');
      
      halfspaces{i}=name;
     % disp(halfspaces{i})
   end
end

read_a_line = fgetl(input_file_id);
if(~strcmp(read_a_line,'MACHINES'))
   error('Incorrect intermediate file: MACHINES expected')
end
read_a_line = fgetl(input_file_id);

add_block('template/Mux', [sys_name '/Mux']);
num_m = str2num(read_a_line); %Number of machines in the model
set_param([sys_name '/Mux'],'Inputs', num2str(num_m));
set_param( [sys_name '/Mux'], 'Position', [ 500, 100, 520,num_m*100+50]); 
%3 - create FSMs corresponding to machines
read_a_line = fgetl(input_file_id);
if(~strcmp(read_a_line,'machine'))
  error('Incorrect intermediate file: machine expected')
end
for m=1:num_m
   machine_name = fgetl(input_file_id);
   machines(m).num_input_events = 0;
   machines(m).num_output_events = 0;
%    machines(m).ievents = num_input_events;
%    machines(m).oevents = num_input_events;
   
   %Store this stuff in an array of 'machine' structures
   machines(m).name = machine_name;
   
 %  disp(['MACHINE' machine_name])
   
   add_block('template/fsm', [sys_name '/' machine_name]);
   set_param( [sys_name '/' machine_name], 'Position', [ 400 ,m*100 , 450, m*100+50]); 
%   [locations,dummy]=find_node(machines{m}.value,'location_list');
%   [halfspaces,dummy]=find_node(parse_tree.value,'half_space_list');
   
   read_a_line = fgetl(input_file_id);
   num_of_states=str2num(read_a_line);
   read_a_line = fgetl(input_file_id);
   machines(m).ini_loc=read_a_line; %initial location
   
   read_a_line = fgetl(input_file_id);
   
   if(strcmp(read_a_line,'I_EVENTS'))
      read_a_line = fgetl(input_file_id);
      num_input_events = str2num(read_a_line);
      
      machines(m).num_input_events = num_input_events;
      
%       if(num_input_events > 1)
%          % Add a MUX for the input events   
%          add_block('template/Mux', [sys_name '/IMux']);
%          set_param( [sys_name '/IMux'],'Inputs', num_input_events);
%          set_param( [sys_name '/IMux'], 'Position', [ 500, 50, 550,num_m*100]); %ADJUST POSITION LATER
%       end
      
      read_a_line = fgetl(input_file_id);
      idx=1;
      while(~strcmp(read_a_line,'ENDI_EVENTS'))
         machines(m).ievents{idx} = read_a_line;
         read_a_line = fgetl(input_file_id);
         idx = idx + 1;
      end
      
      read_a_line = fgetl(input_file_id);
      if(~strcmp(read_a_line,'O_EVENTS'))
         error('Incorrect intermediate file: "O_EVENTS" expected')   
      end
      
      read_a_line = fgetl(input_file_id);
      num_output_events = str2num(read_a_line);
      
      machines(m).num_output_events = num_output_events;
      read_a_line = fgetl(input_file_id);
      idx=1;
      while(~strcmp(read_a_line,'ENDO_EVENTS'))
         machines(m).oevents{idx} = read_a_line;
         read_a_line = fgetl(input_file_id);
         idx = idx + 1;
      end
      read_a_line=fgetl(input_file_id);
   end   
   
   if(~strcmp(read_a_line,'state'))
      error('Incorrect intermediate file: "state" expected')
   end
   
%   disp(['MACHINE ' num2str(num_of_states)])
   num_trans={};
   locations={};
   transitions={,};
   for j=1: num_of_states
      i=num_of_states - j + 1;
      
      num_trans{i}=0;
      state_name = fgetl(input_file_id);
      locations{i} = state_name;
   
      read_a_line = fgetl(input_file_id);
      idx=0;
      while(~strcmp(read_a_line,'state'))&(~strcmp(read_a_line,'machine'))&(~(read_a_line==-1))
         idx=idx+1;
         transitions{i,idx} = read_a_line;
         num_trans{i} = num_trans{i}+1;
         read_a_line = fgetl(input_file_id);
      end
  %    disp(['MACHINEtransitions' num2str(i) ' ' num2str(num_trans{i})]);
   end
   
%    disp(['Calling "hspaces", parameters:'])
%    locations
%    transitions
%    num_trans
%    halfspaces
%    disp('length(locations)')
%    length(locations)
   hspaces = find_input_data(halfspaces, locations, transitions, num_trans);
%    disp('Before create_fsm')
%    hspaces
   reset_flag=0;
%   disp(['*' machine_name '*'])
   reset_flag = create_fsm(sys_name,hspaces,machine_name, input_file_id, ...
      locations, transitions, num_of_states, num_trans, machines(m), m);
   
   machines(m).reset_flag = reset_flag;
   for i=1:length(hspaces)
      name=hspaces{i}; %get the name of Polyhedral Threshold
%       name
%       machine_name
      add_line(sys_name, [name '/1'], [machine_name '/' num2str(i)], 'autorouting', 'on');
   end
   add_line(sys_name, [machine_name '/1'], ['Mux/' num2str(m)], 'autorouting', 'on');
   if(reset_flag==1)
      add_line(sys_name, [machine_name '/2'], ['SCSB/2'], 'autorouting', 'on');
   end
end

%NEW CODE FOR EVENT SYNCHRONIZATION: This is currently untested
for m=1:num_m
   if(machines(m).num_input_events > 1)
      add_block('template/Mux', [sys_name '/Mux' machines(m).name]);
      set_param([sys_name '/Mux' machines(m).name],'Inputs', num2str(machines(m).num_input_events));
      set_param( [sys_name '/Mux' machines(m).name], 'Position', [ 400 ,m*100-40, 410, m*100-10]);  
      % Adjust position later
      
      muxname = strcat('Mux',machines(m).name);
 %     disp(['*' machines(m).name '*'])
      
      mac_name = strcat(machines(m).name, '');
      add_line(sys_name, [muxname '/1'], [machines(m).name '/Trigger'], 'autorouting', 'on');
   end
end

for m=1:num_m
   %for each output event put lines to the respective MUX. 
   for n=1:machines(m).num_output_events
      outeve = machines(m).oevents{n};
      for  p=1:num_m
         %search for which machines take this event as input and add lines.
         i=search_event(outeve,machines(p));
         if(i>0)
            muxname = strcat('Mux',machines(p).name);
            if(machines(p).num_input_events > 1)
               if((machines(m).reset_flag)==1)
                  add_line(sys_name, [machines(m).name '/' num2str(2+n)], [muxname '/' num2str(i)], 'autorouting', 'on');
               else
                  add_line(sys_name, [machines(m).name '/' num2str(1+n)], [muxname '/' num2str(i)], 'autorouting', 'on');
               end
            else
               if((machines(m).reset_flag)==1)
                  add_line(sys_name, [machines(m).name '/' num2str(2+n)], [machines(p).name '/Trigger'], 'autorouting', 'on');
               else
                  add_line(sys_name, [machines(m).name '/' num2str(1+n)], [machines(p).name '/Trigger'], 'autorouting', 'on');
               end
            end
         end
      end
   end
end

while ~(read_a_line==-1)
    read_a_line=fgetl(input_file_id);
end

add_line(sys_name, 'Mux/1', 'SCSB/1', 'autorouting', 'on');
close_system('template');
save_system(sys_name);
close_system(sys_name);

disp([' '])
disp([' '])
disp(['Model ' ascii_intr_file '.mdl created. ' ascii_intr_file '_setup.m must be run first.'])

return

%===============================================================================
function i = search_event(out_eve,machine)
   i = 0;
   for j=1:machine.num_input_events
      if(strcmp(out_eve,machine.ievents{j}))
         i = j;
         break;
      end
   end
return

%-------------------------------------------------------------------------------
function reset_flag = create_fsm(sys_name, halfspaces, machine_name, ... 
   input_file_id, locations, transitions, num_of_states, num_trans, machine, mac_idx)
% create the fsm for the system

rt = sfroot;
m = rt.find('-isa', 'Stateflow.Machine','-and','Name',sys_name);
chart = m.find('-isa','Stateflow.Chart', '-and', 'Name', machine_name);
% calculate the position of that state
states_per_line=ceil(num_of_states/4);
state={};
ini=1;

for i=1: num_of_states
   line=ceil(i/states_per_line);

   %calculate the position of that state
   switch line
   case(1)
      position=[200*i+100, 100, 100, 40];
   case(2)
      position=[200*states_per_line+100, 200*(i-states_per_line)+100,100,40]; 
   case(3)
      position=[200*(3*states_per_line-i)+100, 200*states_per_line+100, 100,40];
   case(4)
      position=[100, 200*(4*states_per_line-i)+100,100,40];
   end

   state_name = locations{i};
   state{i} = Stateflow.State(chart);
   state{i}.Name = locations{i};
   if(strcmp(locations{i},machine.ini_loc))
       ini=i;
   end
   state{i}.Position= position;
   state{i}.LabelString = [state_name ' / q=' num2str(i) ';'];
end
%create input
for i=1:length(halfspaces)
   name=halfspaces{i}; %get the name of Polyhedral Threshold
   in_data= Stateflow.Data(chart);
   in_data.Name=name;
   in_data.Scope= 'INPUT_DATA';
   in_data.DataType= 'boolean';
end
%create output
out_data= Stateflow.Data(chart);
out_data.Name='q';
out_data.Scope= 'OUTPUT_DATA';
out_data.DataType= 'int8';
%create transitions

reset_flag=0;

%disp('CREATING TRANSITIONS');
for i=1:num_of_states
%   disp(['creating transitions loop 1, ' locations{i} ' ' num2str(num_trans{i})])
   for j=1:num_trans{i}
%      disp(transitions{i,j})
      [ievent trans] = find_ievent(transitions{i,j});
      
      [oevent trans] = find_oevent(trans);
      
      [reset trans] = find_target(trans);
%      disp(['reset=' reset ' ' trans]);
      if(strcmp('reset',reset))
         flag=1;
         reset_flag=1;
         [target trans]=find_target(trans);
%         disp(target)
      else 
         flag=0;
         target=reset;
      end
      
      for k=1:num_of_states
         if strcmp(state{k}.Name,target)
            dest=k;
               break;
         end
      end
      [guard]=find_guard(trans);
 %     disp(guard)
      
%      disp(['source ' locations{i} ' dest ' state{k}])
      t = Stateflow.Transition(chart);
      t.Source = state{i};
      t.Destination = state{k};
 
      if(flag==1)
         if(~strcmp(oevent,''))
            %output event not null for this transition
            if(~strcmp(ievent,''))
               t.LabelString=[ guard '/ reset;' oevent];   
            else
               t.LabelString=['[' guard ']/ reset;' oevent];
            end
         else
            if(~strcmp(ievent,''))
               t.LabelString=[ guard '/ reset'];
            else
               t.LabelString=[ '[' guard ']/ reset'];
            end
         end
      else 
         if(~strcmp(oevent,''))
            %output event not null for this transition
            if(~strcmp(ievent,''))
               t.LabelString=[ guard '/ ' oevent];  
            else
               t.LabelString=['[' guard ']/ ' oevent];  
            end
         else
            if(~strcmp(ievent,''))
               t.LabelString=[ guard ];
            else
               t.LabelString=['[' guard ']'];
            end
         end
      end
   end
end
%  create an output reset event
if(reset_flag==1)
   out_event= Stateflow.Event(chart);
   out_event.Name='reset';
   out_event.Scope= 'OUTPUT_EVENT';
   out_event.Trigger= 'RISING_EDGE_EVENT';
end

% NEW CODE FOR EVENT SYNCHRONIZATION
for i=1:machine.num_input_events
   in_event = Stateflow.Event(chart);
   in_event.Name= machine.ievents{i};
   in_event.Scope= 'INPUT_EVENT';
   in_event.Trigger= 'RISING_EDGE_EVENT';
end

for i=1:machine.num_output_events
   out_event= Stateflow.Event(chart);
   out_event.Name=machine.oevents{i};
   out_event.Scope= 'OUTPUT_EVENT';
   out_event.Trigger= 'RISING_EDGE_EVENT';
end

%create start 
t = Stateflow.Transition(chart);
%t.Destination = state{1};
t.Destination = state{ini};
t.DestinationOClock = 12.;
return

%-----------------------------------------------------------------------------------------------
function halfspace_used=find_input_data(halfspaces, locations, trans, num_trans)
halfspace_nos=[];
halfspace_used={};
idx=1;
for loc=1:length(locations)
    location=locations{loc};
%     disp(['in find_input_data'])
%     disp('length num_trans')
%     length(num_trans)
%    disp(['loc ' num2str(loc)])
    if(~(num_trans{loc}==0))
       for t=1:num_trans{loc}
    %      guard=find_guard(trans{loc,t});
     % New Code
          [ievent trns] = find_ievent(trans{loc,t});
      
          [oevent trns] = find_oevent(trns);
      
          [reset trns] = find_target(trns);
%          disp(['reset=' reset ' ' trns]);
          if(strcmp('reset',reset))
             flag=1;
             [target trns]=find_target(trns);
%             disp(target)
          else 
             flag=0;
             target=reset;
          end
          guard=find_guard(trns);
          
          while(~strcmp(guard,''))
             [hspace,guard]=get_halfspace(guard);
             
             % Now check whether 'guard' is actually a halfspace
             is_halfspace=0;
             for k=1:length(halfspaces)
                if(strcmp(hspace,halfspaces{k}))
                   is_halfspace=1;
                end
             end
             
             if(is_halfspace==1)
                flag=0;
                for j=1:idx-1
                   if(strcmp(hspace,halfspace_used{j}))
                      flag = 1;
                      break;
                   end
                end
                if(flag==0)
                   halfspace_used{idx}=hspace;
                   idx=idx+1;
                end
             end
          end
       end
    end
end

% disp(['in find_input_data']);
% halfspace_used
return

%----------------------------------------------------------------------------------------------
function initial_point=get_initial_point(sys_name)
eval([sys_name '_setup']); %setup the file in base
initial_set_linearcon= evalin('base','initial_set');
[CE,dE, CI,dI] = linearcon_data(initial_set_linearcon{1});
C = [CE;CI];
d = [dE;dI];
v1 = C\d;
initial_point=mat2str(v1);
return

%----------------------------------------------------------------------------------------------
function [ievent, trans] = find_ievent(trans)
   [ievent t] = strtok(trans,'+');
   
   if(strcmp(ievent,trans))
      %there is no oevent
      ievent='';
   else
      word=[];
      trans = t;
      
      temp=cellstr(trans);
      temp1=char(temp{1});
      for i=2:length(temp1)
         word=strcat(word,temp1(i));
      end
   
      trans = word;
   end
return

%----------------------------------------------------------------------------------------------
function [oevent, trans] = find_oevent(trans)
   [oevent t] = strtok(trans,'*');
   
   if(strcmp(oevent,trans))
      %there is no oevent
      oevent='';
   else
      word=[];
      trans = t;    
      
      temp=cellstr(trans);
      temp1=char(temp{1});
      for i=2:length(temp1)
         word=strcat(word,temp1(i));
      end
   
      trans = word;
   end
return

%----------------------------------------------------------------------------------------------
function [target, trans] = find_target(trans)
   [target trans] = strtok(trans,';');
   word=[];
   
   temp=cellstr(trans);
   temp1=char(temp{1});
   for i=2:length(temp1)
      word=strcat(word,temp1(i));
   end
   
   trans = word;
return


%----------------------------------------------------------------------------------------------
function guard = find_guard(trans)
%    [dummy guard] = strtok(trans,';');
%    word=[];
%    
%    temp=cellstr(guard);
%    temp1=char(temp{1});
%    for i=1:length(temp1)
%       word=strcat(word,temp1(i));
%    end
%    
%    guard = word;
   guard = trans;
return

%-----------------------------------------------------------------------------------------------
function [hspace,nguard]=get_halfspace(guard)
   word=[];
   nguard1=[];
   
   temp=cellstr(guard);
   temp1=char(temp{1});
   if(length(temp1)<1)
      disp('length temp1 < 1')
      disp(length(temp1))
   end
   
   for j=1:length(temp1)
      if(~((temp1(j)=='(') | (temp1(j)==')') | (temp1(j)=='~') | (temp1(j)=='&') | (temp1(j)=='|') | (temp1(j)==' ')))
         break
      end
   end
   
   for i=j:length(temp1)
      if((temp1(i)=='(') | (temp1(i)==')') | (temp1(i)=='~') | (temp1(i)=='&') | (temp1(i)=='|') | (temp1(i)==' '))
         break
      else
         word=strcat(word,temp1(i));
      end
   end
   
   if(i==length(temp1))
      hspace=word;
   else
      while((temp1(i)=='(')|(temp1(i)==')')|(temp1(i)=='~')|(temp1(i)=='&')|(temp1(i)=='|')|(temp1(i)==' '))
         i = i+ 1;
         if(i==length(temp1))
            break;
         end
      end
      while(i<=length(temp1))
         nguard1=strcat(nguard1,temp1(i));
         i=i+1;
      end
   end
   
%    disp(['get_halfspace'])
%    disp('guard');
%    guard
%    disp('returning ...')
   
   hspace=char(word);
%   hspace
   nguard=char(nguard1);
   
return